import pytest
from mat3ra.standata.applications import ApplicationStandata
from mat3ra.standata.subworkflows import SubworkflowStandata
from mat3ra.standata.workflows import WorkflowStandata

from mat3ra.wode import Subworkflow, Unit, Workflow

WORKFLOW_STANDATA = WorkflowStandata()
SUBWORKFLOW_STANDATA = SubworkflowStandata()
APPLICATION_STANDATA = ApplicationStandata()

WORKFLOW_NAME = WORKFLOW_STANDATA.get_by_name_first_match("band_gap")["name"]
SUBWORKFLOW_NAME = SUBWORKFLOW_STANDATA.get_by_name_first_match("pw_scf")["name"]
DEFAULT_WF_NAME = WORKFLOW_STANDATA.get_default()["name"]

APPLICATION_ESPRESSO = APPLICATION_STANDATA.get_by_name_first_match("espresso")["name"]
APPLICATION_VASP = APPLICATION_STANDATA.get_by_name_first_match("vasp")["name"]
APPLICATION_PYTHON = APPLICATION_STANDATA.get_by_name_first_match("python")["name"]
RELAXATION_NAME = SUBWORKFLOW_STANDATA.get_relaxation_by_application(APPLICATION_ESPRESSO)["name"]

UNIT_CONFIG = {
    "type": "execution",
    "name": "pw_scf",
    "flowchartId": "unit-flowchart-id",
    "head": True,
}


def test_creation():
    wf = Workflow(name=WORKFLOW_NAME)
    assert wf.name == WORKFLOW_NAME


def test_subworkflows():
    sw = Subworkflow(name=SUBWORKFLOW_NAME)
    wf = Workflow(name=WORKFLOW_NAME, subworkflows=[sw])
    assert len(wf.subworkflows) == 1
    assert wf.subworkflows[0].name == SUBWORKFLOW_NAME


def test_with_units():
    unit = Unit(**UNIT_CONFIG)
    wf = Workflow(name=WORKFLOW_NAME, units=[unit])
    assert len(wf.units) == 1
    assert wf.units[0].name == UNIT_CONFIG["name"]


def test_id_generation():
    wf1 = Workflow(name=WORKFLOW_NAME)
    wf2 = Workflow(name=WORKFLOW_NAME)
    assert wf1.id != wf2.id


def test_to_dict():
    wf = Workflow(name=WORKFLOW_NAME)
    data = wf.to_dict()
    assert data["name"] == WORKFLOW_NAME


def test_add_subworkflow():
    wf = Workflow(name=WORKFLOW_NAME)
    sw = Subworkflow(name=SUBWORKFLOW_NAME)
    wf.add_subworkflow(sw)
    assert len(wf.subworkflows) == 1
    assert wf.subworkflows[0].name == SUBWORKFLOW_NAME
    assert len(wf.units) == 1
    assert wf.units[0].name == SUBWORKFLOW_NAME
    assert wf.units[0].type == "subworkflow"


@pytest.mark.parametrize(
    "application,has_relaxation",
    [
        (APPLICATION_ESPRESSO, True),
        (APPLICATION_VASP, True),
        (APPLICATION_PYTHON, False),
    ],
)
def test_get_relaxation_subworkflow(application, has_relaxation):
    workflows = WORKFLOW_STANDATA.get_by_categories(application, DEFAULT_WF_NAME)
    if not workflows:
        pytest.skip(f"No {DEFAULT_WF_NAME} workflow found for {application}")

    workflow_config = workflows[0]
    wf = Workflow(**workflow_config)

    result = wf.relaxation_subworkflow
    if has_relaxation:
        assert result is not None
        assert result.name == RELAXATION_NAME
        assert hasattr(result, "name")
    else:
        assert result is None


@pytest.mark.parametrize(
    "application",
    [APPLICATION_ESPRESSO, APPLICATION_VASP],
)
def test_add_relaxation(application):
    workflows = WORKFLOW_STANDATA.get_by_categories(application, DEFAULT_WF_NAME)
    if not workflows:
        pytest.skip(f"No {DEFAULT_WF_NAME} workflow found for {application}")

    workflow_config = workflows[0]
    wf = Workflow(**workflow_config)

    initial_subworkflow_count = len(wf.subworkflows)
    assert not wf.has_relaxation

    wf.add_relaxation()

    assert wf.has_relaxation
    assert len(wf.subworkflows) == initial_subworkflow_count + 1
    assert wf.subworkflows[0].name == wf.relaxation_subworkflow.name


@pytest.mark.parametrize(
    "application",
    [APPLICATION_ESPRESSO, APPLICATION_VASP],
)
def test_remove_relaxation(application):
    workflows = WORKFLOW_STANDATA.get_by_categories(application, DEFAULT_WF_NAME)
    if not workflows:
        pytest.skip(f"No {DEFAULT_WF_NAME} workflow found for {application}")

    workflow_config = workflows[0]
    wf = Workflow(**workflow_config)

    wf.add_relaxation()
    assert wf.has_relaxation
    initial_subworkflow_count = len(wf.subworkflows)

    wf.remove_relaxation()

    assert not wf.has_relaxation
    assert len(wf.subworkflows) == initial_subworkflow_count - 1


@pytest.mark.parametrize(
    "method",
    [
        "only_new_unit",
        "with_unit_instance",
        "with_flowchart_id",
    ],
)
def test_set_unit(method):
    workflows = WORKFLOW_STANDATA.get_by_categories(APPLICATION_ESPRESSO, DEFAULT_WF_NAME)
    if not workflows:
        pytest.skip(f"No {DEFAULT_WF_NAME} workflow found for {APPLICATION_ESPRESSO}")

    workflow_config = workflows[0]
    wf = Workflow(**workflow_config)

    wf.add_relaxation()

    unit_to_modify = wf.get_unit_by_name(name_regex="relax")
    assert unit_to_modify is not None

    new_context = {"test_key": "test_value", "another_key": 42}
    unit_to_modify.add_context(new_context)

    if method == "only_new_unit":
        success = wf.set_unit(unit_to_modify)
    elif method == "with_unit_instance":
        original_unit = wf.get_unit_by_name(name_regex="relax")
        success = wf.set_unit(unit_to_modify, unit=original_unit)
    elif method == "with_flowchart_id":
        flowchart_id = unit_to_modify.flowchartId
        success = wf.set_unit(unit_to_modify, unit_flowchart_id=flowchart_id)

    assert success is True

    updated_unit = wf.get_unit_by_name(name_regex="relax")
    assert "test_key" in updated_unit.context
    assert "another_key" in updated_unit.context
    assert updated_unit.context["test_key"] == "test_value"
    assert updated_unit.context["another_key"] == 42



import pytest
from mat3ra.wode.units import Unit
from mat3ra.wode.workflows import Workflow

UNIT_1_NAME = "unit_1"
UNIT_2_NAME = "unit_2"
UNIT_3_NAME = "unit_3"
UNIT_TAG = "test_tag"
FLOWCHART_ID_1 = "flowchart-id-1"
FLOWCHART_ID_2 = "flowchart-id-2"
FLOWCHART_ID_3 = "flowchart-id-3"

UNIT_CONFIG_1 = {
    "type": "execution",
    "name": UNIT_1_NAME,
    "flowchartId": FLOWCHART_ID_1,
}

UNIT_CONFIG_2 = {
    "type": "execution",
    "name": UNIT_2_NAME,
    "flowchartId": FLOWCHART_ID_2,
}

UNIT_CONFIG_3 = {
    "type": "execution",
    "name": UNIT_3_NAME,
    "flowchartId": FLOWCHART_ID_3,
    "tags": [UNIT_TAG],
}


@pytest.fixture
def workflow():
    """Create a workflow instance for testing FlowchartUnitsManager methods."""
    return Workflow(name="Test Workflow")


@pytest.fixture
def unit_1():
    return Unit(**UNIT_CONFIG_1)


@pytest.fixture
def unit_2():
    return Unit(**UNIT_CONFIG_2)


@pytest.fixture
def unit_3():
    return Unit(**UNIT_CONFIG_3)


def test_set_units(workflow, unit_1, unit_2):
    units = [unit_1, unit_2]
    workflow.set_units(units)
    assert len(workflow.units) == 2
    assert workflow.units[0].flowchartId == FLOWCHART_ID_1
    assert workflow.units[1].flowchartId == FLOWCHART_ID_2


def test_get_unit(workflow, unit_1, unit_2):
    workflow.set_units([unit_1, unit_2])
    found_unit = workflow.get_unit(FLOWCHART_ID_1)
    assert found_unit is not None
    assert found_unit.flowchartId == FLOWCHART_ID_1
    assert found_unit.name == UNIT_1_NAME


def test_find_unit_by_id(workflow, unit_1, unit_2):
    workflow.set_units([unit_1, unit_2])
    found_unit = workflow.find_unit_by_id(unit_1.id)
    assert found_unit is not None
    assert found_unit.id == unit_1.id


def test_find_unit_with_tag(workflow, unit_3):
    workflow.set_units([unit_3])
    found_unit = workflow.find_unit_with_tag(UNIT_TAG)
    assert found_unit is not None
    assert UNIT_TAG in found_unit.tags


@pytest.mark.parametrize(
    "search_name,expected_name",
    [
        (UNIT_1_NAME, UNIT_1_NAME),
        (UNIT_1_NAME.upper(), UNIT_1_NAME),  # Case insensitive
        (UNIT_2_NAME, UNIT_2_NAME),
    ],
)
def test_get_unit_by_name(workflow, unit_1, unit_2, search_name, expected_name):
    workflow.set_units([unit_1, unit_2])
    found_unit = workflow.get_unit_by_name(name=search_name)
    assert found_unit is not None
    assert found_unit.name == expected_name


def test_get_unit_by_name_regex(workflow, unit_1, unit_2):
    workflow.set_units([unit_1, unit_2])
    found_unit = workflow.get_unit_by_name(name_regex=r"unit_\d")
    assert found_unit is not None
    assert found_unit.name in [UNIT_1_NAME, UNIT_2_NAME]


def test_set_units_head(workflow, unit_1, unit_2, unit_3):
    units = [unit_1, unit_2, unit_3]
    result = workflow.set_units_head(units)
    assert result[0].head is True
    assert result[1].head is False
    assert result[2].head is False


def test_set_next_links(workflow, unit_1, unit_2, unit_3):
    units = [unit_1, unit_2, unit_3]
    result = workflow.set_next_links(units)
    assert result[0].next == FLOWCHART_ID_2
    assert result[1].next == FLOWCHART_ID_3
    assert result[2].next is None or result[2].next == ""


def test_clear_link_to_unit(workflow, unit_1, unit_2):
    unit_1.next = FLOWCHART_ID_2
    workflow.set_units([unit_1, unit_2])
    workflow._clear_link_to_unit(FLOWCHART_ID_2)
    assert unit_1.next is None


def test_add_unit(workflow, unit_1, unit_2):
    workflow.add_unit(unit_1)
    assert len(workflow.units) == 1
    assert workflow.units[0].head is True
    assert workflow.units[0].flowchartId == FLOWCHART_ID_1

    workflow.add_unit(unit_2)
    assert len(workflow.units) == 2
    assert workflow.units[0].head is True
    assert workflow.units[1].head is False
    assert workflow.units[0].next == FLOWCHART_ID_2


@pytest.mark.parametrize(
    "head,expected_order",
    [
        (True, [FLOWCHART_ID_3, FLOWCHART_ID_1, FLOWCHART_ID_2]),
        (False, [FLOWCHART_ID_1, FLOWCHART_ID_2, FLOWCHART_ID_3]),
    ],
)
def test_add_unit_head_parameter(workflow, unit_1, unit_2, unit_3, head, expected_order):
    workflow.add_unit(unit_1)
    workflow.add_unit(unit_2)
    workflow.add_unit(unit_3, head=head)

    actual_order = [u.flowchartId for u in workflow.units]
    assert actual_order == expected_order


def test_remove_unit(workflow, unit_1, unit_2, unit_3):
    workflow.set_units([unit_1, unit_2, unit_3])
    workflow.remove_unit(FLOWCHART_ID_2)

    assert len(workflow.units) == 2
    assert workflow.units[0].flowchartId == FLOWCHART_ID_1
    assert workflow.units[1].flowchartId == FLOWCHART_ID_3
    assert workflow.units[0].next == FLOWCHART_ID_3


def test_replace_unit(workflow, unit_1, unit_2):
    workflow.set_units([unit_1])
    workflow.replace_unit(0, unit_2)

    assert len(workflow.units) == 1
    assert workflow.units[0].flowchartId == FLOWCHART_ID_2


@pytest.mark.parametrize(
    "provide_unit,provide_id,should_succeed",
    [
        (True, False, True),  # Provide unit instance
        (False, True, True),  # Provide flowchart_id
        (False, False, True),  # Provide neither (use new_unit.flowchartId)
    ],
)
def test_set_unit(workflow, unit_1, unit_2, provide_unit, provide_id, should_succeed):
    workflow.set_units([unit_1])

    # Create new unit with same flowchart_id
    new_unit = Unit(type="execution", name="Updated Unit", flowchartId=FLOWCHART_ID_1)

    if provide_unit:
        result = workflow.set_unit(new_unit, unit=unit_1)
    elif provide_id:
        result = workflow.set_unit(new_unit, unit_flowchart_id=FLOWCHART_ID_1)
    else:
        result = workflow.set_unit(new_unit)

    assert result is should_succeed
    if should_succeed:
        assert workflow.units[0].name == "Updated Unit"


def test_set_unit_replaces_itself(workflow):
    unit = Unit(type="execution", name="Original", flowchartId=FLOWCHART_ID_1)
    workflow.add_unit(unit)

    # Modify the unit outside
    unit.name = "Modified"

    # Replace it with itself using flowchartId
    result = workflow.set_unit(unit)

    assert result is True
    assert workflow.units[0].name == "Modified"



from typing import List

from mat3ra.ade import Application, Executable, Flavor
from mat3ra.esse.models.workflow.unit.execution import ExecutionUnitSchemaBase
from pydantic import Field

from .unit import Unit


class ExecutionUnit(Unit, ExecutionUnitSchemaBase):
    executable: Executable = None
    flavor: Flavor = None
    application: Application = None
    input: List = Field(default_factory=list)



import pytest
from mat3ra.wode.context.providers import PointsGridDataProvider
from mat3ra.esse.models.context_providers_directory.points_grid_data_provider import GridMetricType

# Test data constants
DIMENSIONS_DEFAULT = [1, 1, 1]
DIMENSIONS_CUSTOM = [1, 2, 3]
SHIFTS_DEFAULT = [0.0, 0.0, 0.0]
SHIFTS_CUSTOM = [0.5, 0.5, 0.5]
DIVISOR_DEFAULT = 1
DIVISOR_CUSTOM = 2
GRID_METRIC_TYPE_DEFAULT = GridMetricType.KPPRA

# Expected data structures
KGRID_DATA = {
    "kgrid": {
        "dimensions": DIMENSIONS_CUSTOM,
        "shifts": SHIFTS_DEFAULT,
        "divisor": DIVISOR_DEFAULT,
        "gridMetricType": GRID_METRIC_TYPE_DEFAULT,
    },
    "isKgridEdited": True,
}


@pytest.mark.parametrize(
    "init_params,expected_dimensions,expected_shifts,expected_divisor",
    [
        (
            {"dimensions": DIMENSIONS_CUSTOM},
            DIMENSIONS_CUSTOM,
            SHIFTS_DEFAULT,
            DIVISOR_DEFAULT,
        ),
    ],
)
def test_points_grid_data_provider_initialization(init_params, expected_dimensions, expected_shifts, expected_divisor):
    kgrid_context_provider_relax = PointsGridDataProvider(**init_params)

    assert kgrid_context_provider_relax.dimensions == expected_dimensions
    assert kgrid_context_provider_relax.shifts == expected_shifts
    assert kgrid_context_provider_relax.divisor == expected_divisor


@pytest.mark.parametrize(
    "init_params,expected_data",
    [
        (
                {"dimensions": DIMENSIONS_CUSTOM},
                KGRID_DATA,
        ),
    ],
)
def test_points_grid_data_provider_get_data(init_params, expected_data):
    kgrid_context_provider = PointsGridDataProvider(**init_params)
    actual_data = kgrid_context_provider.get_data()
    assert actual_data == expected_data["kgrid"]



@pytest.mark.parametrize(
    "init_params,expected_data",
    [
        (
            {"dimensions": DIMENSIONS_CUSTOM, "is_edited": True},
            KGRID_DATA,
        ),
    ],
)
def test_points_grid_data_provider_yield_data(init_params, expected_data):
    kgrid_context_provider = PointsGridDataProvider(**init_params)
    actual_data = kgrid_context_provider.yield_data()
    assert actual_data == expected_data




from typing import List, Optional

from mat3ra.ade.application import Application
from mat3ra.code.entity import InMemoryEntitySnakeCase
from mat3ra.esse.models.workflow.subworkflow import Subworkflow as SubworkflowSchema
from mat3ra.mode.method import Method
from mat3ra.mode.model import Model
from mat3ra.utils.uuid import get_uuid
from pydantic import Field

from ..mixins import FlowchartUnitsManager
from ..units import Unit


class Subworkflow(SubworkflowSchema, InMemoryEntitySnakeCase, FlowchartUnitsManager):
    """
    Subworkflow class representing a logical collection of workflow units.

    Attributes:
        name: Name of the subworkflow
        application: Application configuration
        model: Model configuration
        units: List of units in the subworkflow
        properties: List of properties extracted by the subworkflow
    """

    id: str = Field(default_factory=get_uuid, alias="_id")
    application: Application = Field(
        default_factory=lambda: Application(name="", version="", build="", shortName="", summary="")
    )
    model: Model = Field(default_factory=lambda: Model(type="", subtype="", method=Method(type="", subtype="")))
    units: List[Unit] = Field(default_factory=list)

    @classmethod
    def from_arguments(
        cls,
        application: Application,
        model: Model,
        method: Method,
        name: str,
        units: Optional[List] = None,
        config: Optional[dict] = None,
    ) -> "Subworkflow":
        if units is None:
            units = []
        if config is None:
            config = {}

        model.method = method
        return cls(name=name, application=application, model=model, units=units, **config)

    @property
    def method_data(self):
        return self.model.method.data

    def get_as_unit(self) -> Unit:
        return Unit(type="subworkflow", id=self.id, name=self.name)
